"""
Copyright (C) 2022-2024 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

import re
import uuid


def flatten_field(field):
    if isinstance(field.valueType, dict):
        # Handle nested types
        for nested_field_name, nested_field_value_type in field.valueType.items():
            yield f"{field.name}.{nested_field_name}", nested_field_value_type
    else:
        yield field.name, field.valueType


def is_valid_uuid(uuid_to_test, version=4):
    """
    Check if uuid_to_test is a valid UUID.

    Parameters:
    * uuid_to_test: str - The string to test as a UUID.
    * version: int - UUID version (default is 4).

    Returns:
    * `True` if uuid_to_test is a valid UUID; otherwise `False`.
    """
    try:
        uuid_obj = uuid.UUID(uuid_to_test, version=version)
    except ValueError:
        return False
    return str(uuid_obj) == uuid_to_test


def remove_comments(code: str) -> str:
    return re.sub(r"/\*.*?\*/", "", code, flags=re.DOTALL)


def snake_to_camel(snake_str):
    components = snake_str.split("_")
    return "".join(x.title() for x in components)


def snake_to_lower_camel(s):
    components = s.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


def camel_to_snake(camel_str):
    return "".join(["_" + i.lower() if i.isupper() else i for i in camel_str]).lstrip("_")



"""
Copyright (C) 2022-2024 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

import click

from stellanow_cli.commands.code_generator.events import events_cmd
from stellanow_cli.commands.code_generator.generate import generate_cmd
from stellanow_cli.commands.code_generator.models import models_cmd
from stellanow_cli.commands.code_generator.plan import plan_cmd
from stellanow_cli.core.context import StellaNowContext, pass_stella_context
from stellanow_cli.core.decorators import option_with_config_lookup
from stellanow_cli.core.validators import password_validator, url_validator, uuid_validator
from stellanow_cli.services.code_generator.code_generator import CodeGeneratorService, CodeGeneratorServiceConfig


@click.group(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@option_with_config_lookup(
    "--base_url",
    service_class=CodeGeneratorService,
    callback=url_validator,
    help="Base Service url",
)
@option_with_config_lookup("--username", service_class=CodeGeneratorService, help="Base Service Api key")
@option_with_config_lookup(
    "--password",
    service_class=CodeGeneratorService,
    callback=password_validator,
    hide_input=True,
    help="Base Service Api Secret",
)
@option_with_config_lookup(
    "--organization_id",
    service_class=CodeGeneratorService,
    callback=uuid_validator,
    hide_input=True,
    help="Name of Organization",
)
@pass_stella_context
@click.pass_context
def code_generator_service(
    ctx: click.Context,
    stella_ctx: StellaNowContext,
    base_url: str,
    username: str,
    password: str,
    organization_id: str,
    *args,
    **kwargs
) -> None:
    ctx.obj = CodeGeneratorService(
        CodeGeneratorServiceConfig(
            base_url=base_url,
            username=username,
            password=password,
            organization_id=organization_id,
        )
    )

    stella_ctx.add_service(service=ctx.obj)


code_generator_service.add_command(events_cmd)
code_generator_service.add_command(generate_cmd)
code_generator_service.add_command(plan_cmd)
code_generator_service.add_command(models_cmd)



"""
Copyright (C) 2022-2024 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

from click import ClickException


class StellaNowCLIException(ClickException):
    """Exception raised for errors in the Stella Now CLI."""

    def __init__(self, message, details):
        self.message = message
        self.details = details
        super().__init__(self.message)

    def __str__(self):
        return f"{self.message} {self.details}"


class StellaNowCLILanguageNotSupportedException(StellaNowCLIException):
    """Exception raised for unsupported languages by the CLI."""

    def __init__(self, language):
        super().__init__(f"Code generator for {language} not found.", {})


class StellaNowCLINamespaceNotFoundException(StellaNowCLIException):
    """Exception raised when a namespace is not found in a file."""

    def __init__(self):
        super().__init__(f"No Namespace Found", {})


class StellaNowCLINamespaceNotProvidedException(StellaNowCLIException):
    """Exception raised when a namespace is not provided."""

    def __init__(self):
        super().__init__(f"Namespace must be provided for C# code generation", {})


class StellaNowCLINoEntityAssociatedWithEventException(StellaNowCLIException):
    """Exception raised when an event does not have any associated entities. It is not a valid event for ingestion."""

    def __init__(self):
        super().__init__(f"No Entity Associated With Event", {})



"""
Copyright (C) 2022-2024 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

import glob
import os
from typing import List

import click
from loguru import logger

from stellanow_cli.code_generators import CsharpCodeGenerator
from stellanow_cli.core.helpers import SkippedFile
from stellanow_cli.core.utils.logger_utils import log_summary
from stellanow_cli.core.validators import uuid_validator
from stellanow_cli.services.code_generator.code_generator import CodeGeneratorService, pass_code_generator_service
from stellanow_cli.services.code_generator.tools import process_file


@click.command()
@click.option(
    "--project_id",
    "-p",
    required=True,
    prompt=True,
    callback=uuid_validator,
    help="UUID of the project associated with the organization saved in your configuration file.",
)
@click.option("--input_dir", "-i", default=".", help="The directory to read generated classes from.")
@pass_code_generator_service
def plan(service: CodeGeneratorService, project_id: str, input_dir: str, **kwargs):
    """Compares currently generated classes with the specifications fetched from the API and provides a summary of
    changes."""
    logger.info("Planning...")

    workflow_client = service.create_workflow_client(project_id=project_id)
    generators = {".cs": CsharpCodeGenerator()}
    skipped_files: List[SkippedFile] = []
    files_found = 0

    for filename in glob.iglob(f"{input_dir}/**", recursive=True):
        if os.path.isdir(filename):
            continue

        _, ext = os.path.splitext(filename)
        if ext not in generators:
            continue

        files_found += 1
        generator = generators[ext]

        logger.info(f"==============================\nComparison for file: {filename}")
        reasons = process_file(filename=filename, workflow_client=workflow_client, generator=generator)
        if reasons:
            skipped_files.append(SkippedFile(filename, ", ".join(reasons)))

    if files_found == 0:
        logger.warning("No recognized files found to process. Nothing to plan.")
        return

    log_summary(skipped_files)


plan_cmd = plan



"""
Copyright (C) 2022-2024 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

import difflib
import json
import re
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List

from stellanow_api_internals.datatypes.workflow_mgmt import (
    StellaEntity,
    StellaEventDetailed,
    StellaField,
    StellaFieldType,
    StellaModelDetailed,
    StellaModelField,
)

from stellanow_cli.code_generators.code_generator import CodeGenerator
from stellanow_cli.core.enums import StellaDataStructure
from stellanow_cli.core.utils.string_utils import camel_to_snake, remove_comments, snake_to_camel, snake_to_lower_camel
from stellanow_cli.exceptions.cli_exceptions import (
    StellaNowCLINamespaceNotFoundException,
    StellaNowCLINoEntityAssociatedWithEventException,
)


def field_type_mapping(field: StellaField | StellaModelField, model_details: Dict[str, StellaModelDetailed]) -> str:
    if isinstance(field.fieldType, Dict):
        field_type = StellaFieldType(**field.fieldType)
    else:
        field_type = field.fieldType
    value_type = field_type.value
    if value_type == StellaDataStructure.MODEL:
        return f"{snake_to_camel(model_details[field_type.modelRef].name)}Model" if field_type.modelRef else "object"
    mapping = {
        "Decimal": "decimal",
        "Integer": "int",
        "Boolean": "bool",
        "String": "string",
        "Date": "DateOnly",
        "DateTime": "DateTime",
    }
    return mapping.get(value_type, "object")  # default to object if unknown


def field_format_mapping(value_type: str) -> str:
    mapping = {
        "Decimal": '.ToString("F2", CultureInfo.InvariantCulture)',
        "Integer": ".ToString()",
        "Boolean": ".ToString().ToLower()",
        "String": "",
        "Date": '.ToString("yyyy-MM-dd")',
        "DateTime": '.ToString("yyyy-MM-ddTHH:mm:ss.ffffffZ")',
    }
    return mapping.get(value_type, "")


def escape_reserved_words(word: str) -> str:
    reserved_words = [
        "abstract",
        "as",
        "base",
        "bool",
        "break",
        "byte",
        "case",
        "catch",
        "char",
        "checked",
        "class",
        "const",
        "continue",
        "decimal",
        "default",
        "delegate",
        "do",
        "double",
        "else",
        "enum",
        "event",
        "explicit",
        "extern",
        "false",
        "finally",
        "fixed",
        "float",
        "for",
        "foreach",
        "goto",
        "if",
        "implicit",
        "in",
        "int",
        "interface",
        "internal",
        "is",
        "lock",
        "long",
        "namespace",
        "new",
        "null",
        "object",
        "operator",
        "out",
        "override",
        "params",
        "private",
        "protected",
        "public",
        "readonly",
        "ref",
        "return",
        "sbyte",
        "sealed",
        "short",
        "sizeof",
        "stackalloc",
        "static",
        "string",
        "struct",
        "switch",
        "this",
        "throw",
        "true",
        "try",
        "typeof",
        "uint",
        "ulong",
        "unchecked",
        "unsafe",
        "ushort",
        "using",
        "virtual",
        "void",
        "volatile",
        "while",
    ]

    return f"{word}_" if word in reserved_words else word


@dataclass
class CSharpField:
    original_name: str
    csharp_name: str
    type: str
    is_model: bool

    @classmethod
    def from_stella_field(cls, field: StellaField | StellaModelField, model_details: Dict[str, StellaModelDetailed]):
        field_type = field_type_mapping(field, model_details)
        is_model = field_type.endswith("Model")
        return cls(
            original_name=field.name,
            csharp_name=escape_reserved_words(snake_to_camel(camel_to_snake(field.name))),
            type=field_type,
            is_model=is_model,
        )

    def to_constructor_parameter_declaration(self):
        return f"{self.type} {self.csharp_name}"


@dataclass
class CSharpEntity:
    original_name: str
    csharp_name: str

    @classmethod
    def from_stella_field(cls, entity: StellaEntity):
        return cls(
            original_name=entity.name,
            csharp_name=escape_reserved_words(snake_to_lower_camel(entity.name)) + "Id",
        )

    def to_entity_type_declaration(self):
        return f'new EntityType("{self.original_name}", {self.csharp_name})'

    def to_constructor_parameter_declaration(self):
        return f"string {self.csharp_name}"


@dataclass
class CSharpModel:
    original_name: str
    csharp_name: str


class CsharpCodeGenerator(CodeGenerator):
    @staticmethod
    def generate_message_class(
        event: StellaEventDetailed, model_details: Dict[str, StellaModelDetailed], **kwargs
    ) -> str:
        template = CsharpCodeGenerator.load_template("messages/csharp")

        namespace = kwargs.get("namespace", "StellaNowSDK.Messages")

        fields = [CSharpField.from_stella_field(field=field, model_details=model_details) for field in event.fields]
        entities = [CSharpEntity.from_stella_field(entity) for entity in event.entities]
        if not entities:
            raise StellaNowCLINoEntityAssociatedWithEventException()

        has_model_fields = any(field.type.endswith("Model") for field in fields)

        rendered = template.render(
            className=snake_to_camel(event.name),
            eventName=event.name,
            constructorArguments=", ".join(
                [
                    ", ".join([entity.to_constructor_parameter_declaration() for entity in entities]),
                    ", ".join([field.to_constructor_parameter_declaration() for field in fields]),
                ]
            ),
            entitiesList=", ".join([entity.to_entity_type_declaration() for entity in entities]),
            entities=entities,
            fields=fields,
            namespace=namespace,
            has_model_fields=has_model_fields,
            eventJson=json.dumps(event.model_dump(), indent=4),
            eventId=event.id,
            timestamp=datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
        )

        return rendered

    @staticmethod
    def generate_model_class(
        model: StellaModelDetailed, model_details: Dict[str, StellaModelDetailed], **kwargs
    ) -> str:
        template = CsharpCodeGenerator.load_template("models/csharp")

        namespace = kwargs.get("namespace", "StellaNowSDK.Messages")

        fields = [
            CSharpField.from_stella_field(field=field, model_details=model_details) for field in model.fields.root
        ]

        rendered = template.render(
            className=snake_to_camel(model.name),
            fields=fields,
            namespace=namespace,
            modelJson=json.dumps(model.model_dump(), indent=4),
            modelId=model.id,
            timestamp=datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
        )

        return rendered

    @staticmethod
    def get_file_name_for_event_name(event_name: str) -> str:
        return f"{snake_to_camel(event_name)}Message.cs"

    @staticmethod
    def get_file_name_for_model_name(model_name: str) -> str:
        return f"{snake_to_camel(model_name)}Model.cs"

    @classmethod
    def _get_diff(
        cls, generate_method, item, existing_code: str, models_details: Dict[str, StellaModelDetailed]
    ) -> List[str]:
        namespace_search = re.search(r"namespace (.*);", existing_code)
        if namespace_search is None:
            raise StellaNowCLINamespaceNotFoundException()

        namespace = namespace_search.group(1)

        new_code = generate_method(item, models_details, namespace=namespace)

        existing_code_no_comments = remove_comments(existing_code)
        new_code_no_comments = remove_comments(new_code)

        diff = difflib.unified_diff(
            existing_code_no_comments.splitlines(keepends=True), new_code_no_comments.splitlines(keepends=True)
        )

        return [
            line
            for line in diff
            if line.startswith("- ") or line.startswith("+ ") and not (line.startswith("---") or line.startswith("+++"))
        ]

    @classmethod
    def get_message_diff(
        cls, event: StellaEventDetailed, existing_code: str, models_details: Dict[str, StellaModelDetailed]
    ) -> List[str]:
        return cls._get_diff(cls.generate_message_class, event, existing_code, models_details)

    @classmethod
    def get_model_diff(
        cls, model: StellaModelDetailed, existing_code: str, models_details: Dict[str, StellaModelDetailed]
    ) -> List[str]:
        return cls._get_diff(cls.generate_model_class, model, existing_code, models_details)



"""
Copyright (C) 2022-2024 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

from pathlib import Path
from typing import Any, Callable, Dict, List

from jinja2 import Environment, FileSystemLoader, Template
from stellanow_api_internals.datatypes.workflow_mgmt import StellaEventDetailed, StellaModelDetailed

GenerateMethodType = Callable[[Any, Dict[str, StellaModelDetailed], str], str]


class CodeGenerator:
    @staticmethod
    def load_template(language: str) -> Template:
        # Get the current file path
        current_file_path = Path(__file__).parent

        # Define the relative path to the templates directory
        templates_relative_path = Path("templates")

        # Get the absolute path to the templates directory
        templates_path = current_file_path / templates_relative_path

        # Create a Jinja2 environment with the FileSystemLoader
        env = Environment(loader=FileSystemLoader(templates_path))

        # Get the template
        template = env.get_template(f"{language}.template")

        return template

    @staticmethod
    def generate_message_class(
        event: StellaEventDetailed, model_details: Dict[str, StellaModelDetailed], **kwargs
    ) -> str:
        raise NotImplemented

    @staticmethod
    def generate_model_class(
        model: StellaModelDetailed, model_details: Dict[str, StellaModelDetailed], **kwargs
    ) -> str:
        raise NotImplemented

    @staticmethod
    def get_file_name_for_event_name(event_name: str) -> str:
        raise NotImplemented

    @staticmethod
    def get_file_name_for_model_name(event_name: str) -> str:
        raise NotImplemented

    @staticmethod
    def get_message_diff(
        event: StellaEventDetailed, existing_code: str, models_details: Dict[str, StellaModelDetailed]
    ) -> List[str]:
        raise NotImplemented

    @staticmethod
    def get_model_diff(
        model: StellaModelDetailed, existing_code: str, models_details: Dict[str, StellaModelDetailed]
    ) -> List[str]:
        raise NotImplemented



"""
Copyright (C) 2022-2024 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""
import uuid
from datetime import datetime

import pytest
from loguru import logger
from stellanow_api_internals.datatypes.workflow_mgmt import (
    StellaEventDetailed,
    StellaField,
    StellaFieldType,
    StellaModelDetailed,
    StellaModelField,
    StellaModelFieldType,
    StellaShortEntity,
)

from stellanow_cli.code_generators import CsharpCodeGenerator


@pytest.fixture
def test_event():
    return StellaEventDetailed(
        id="1",
        name="test_event",
        projectId=str(uuid.uuid4()),
        entities=[
            StellaShortEntity(id="", name="entity1"),
            StellaShortEntity(id="", name="entity2"),
        ],
        fields=[
            StellaField(id="", name="field1", fieldType=StellaFieldType(value="Decimal"), required=True),
            StellaField(id="", name="field2", fieldType=StellaFieldType(value="Integer"), required=True),
            StellaField(id="", name="fieldValue3", fieldType=StellaFieldType(value="Boolean"), required=True),
            StellaField(id="", name="field4", fieldType=StellaFieldType(value="String"), required=True),
            StellaField(id="", name="field5", fieldType=StellaFieldType(value="Date"), required=True),
            StellaField(id="", name="field6", fieldType=StellaFieldType(value="DateTime"), required=True),
            StellaField(id="", name="field7", fieldType=StellaFieldType(value="Any"), required=True),
            StellaField(
                id="",
                name="field8",
                fieldType=StellaFieldType(value="Model", modelRef="34a736ed-0b51-415a-ac24-0c6a975e9381"),
                required=True,
            ),
        ],
        isActive=True,
        createdAt=datetime(2022, 1, 1).strftime("%Y-%m-%dT%H:%M:%S"),
        updatedAt=datetime(2022, 1, 2).strftime("%Y-%m-%dT%H:%M:%S"),
        description="Test event",
    )


@pytest.fixture
def test_event_without_model():
    return StellaEventDetailed(
        id="2",
        name="test_event_no_model",
        projectId=str(uuid.uuid4()),
        entities=[
            StellaShortEntity(id="", name="entity1"),
            StellaShortEntity(id="", name="entity2"),
        ],
        fields=[
            StellaField(id="", name="field1", fieldType=StellaFieldType(value="Decimal"), required=True),
            StellaField(id="", name="field2", fieldType=StellaFieldType(value="Integer"), required=True),
            StellaField(id="", name="fieldValue3", fieldType=StellaFieldType(value="Boolean"), required=True),
            StellaField(id="", name="field4", fieldType=StellaFieldType(value="String"), required=True),
            StellaField(id="", name="field5", fieldType=StellaFieldType(value="Date"), required=True),
            StellaField(id="", name="field6", fieldType=StellaFieldType(value="DateTime"), required=True),
            StellaField(id="", name="field7", fieldType=StellaFieldType(value="Any"), required=True),
        ],
        isActive=True,
        createdAt=datetime(2022, 1, 1).strftime("%Y-%m-%dT%H:%M:%S"),
        updatedAt=datetime(2022, 1, 2).strftime("%Y-%m-%dT%H:%M:%S"),
        description="Test event without model",
    )


@pytest.fixture
def model_details():
    return {
        "34a736ed-0b51-415a-ac24-0c6a975e9381": StellaModelDetailed(
            id="34a736ed-0b51-415a-ac24-0c6a975e9381",
            name="TestStructure",
            description="A test model",
            createdAt=datetime(2022, 1, 1).strftime("%Y-%m-%dT%H:%M:%S"),
            updatedAt=datetime(2022, 1, 2).strftime("%Y-%m-%dT%H:%M:%S"),
            fields=[
                StellaModelField(
                    id="fieldA",
                    modelId="34a736ed-0b51-415a-ac24-0c6a975e9381",
                    name="fieldA",
                    fieldType=StellaModelFieldType(value="String"),
                ),
                StellaModelField(
                    id="fieldB",
                    modelId="34a736ed-0b51-415a-ac24-0c6a975e9381",
                    name="fieldB",
                    fieldType=StellaModelFieldType(value="Integer"),
                ),
            ],
        ),
        "66a744ed-0b32-125a-ac24-0c6a525e9381": StellaModelDetailed(
            id="66a744ed-0b32-125a-ac24-0c6a525e9381",
            name="TestNested",
            description="A test model",
            createdAt=datetime(2022, 1, 1).strftime("%Y-%m-%dT%H:%M:%S"),
            updatedAt=datetime(2022, 1, 2).strftime("%Y-%m-%dT%H:%M:%S"),
            fields=[
                StellaModelField(
                    id="fieldC",
                    modelId="66a744ed-0b32-125a-ac24-0c6a525e9381",
                    name="fieldC",
                    fieldType=StellaModelFieldType(value="String"),
                ),
                StellaModelField(
                    id="fieldD",
                    modelId="66a744ed-0b32-125a-ac24-0c6a525e9381",
                    name="fieldD",
                    fieldType=StellaModelFieldType(value="Model", modelRef="87a324ed-0b21-655a-ac24-0c6a755e9381"),
                ),
            ],
        ),
        "87a324ed-0b21-655a-ac24-0c6a755e9381": StellaModelDetailed(
            id="87a324ed-0b21-655a-ac24-0c6a755e9381",
            name="TestAnother",
            description="A test model",
            createdAt=datetime(2022, 1, 1).strftime("%Y-%m-%dT%H:%M:%S"),
            updatedAt=datetime(2022, 1, 2).strftime("%Y-%m-%dT%H:%M:%S"),
            fields=[
                StellaModelField(
                    id="fieldE",
                    modelId="87a324ed-0b21-655a-ac24-0c6a755e9381",
                    name="fieldE",
                    fieldType=StellaModelFieldType(value="String"),
                ),
                StellaModelField(
                    id="fieldF",
                    modelId="87a324ed-0b21-655a-ac24-0c6a755e9381",
                    name="fieldF",
                    fieldType=StellaModelFieldType(value="Integer"),
                ),
            ],
        ),
    }


@pytest.fixture
def model_details_empty():
    return {}


def test_generate_class_handles_all_valueTypes(test_event, model_details):
    # Generate the class
    generated_class = CsharpCodeGenerator.generate_message_class(test_event, model_details)
    logger.info(generated_class)

    # Assertions
    assert "/*\n *  This file is auto-generated by StellaNowCLI. DO NOT EDIT.\n *\n *" in generated_class
    assert "using StellaNowSDK.Messages;" in generated_class
    assert "using StellaNowSDK.Messages.Models;" in generated_class
    assert "namespace StellaNowSDK.Messages;" in generated_class
    assert "[property: Newtonsoft.Json.JsonIgnore] string entity1Id," in generated_class
    assert "[property: Newtonsoft.Json.JsonIgnore] string entity2Id," in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field1")] decimal Field1,' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field2")] int Field2,' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("fieldValue3")] bool FieldValue3,' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field4")] string Field4,' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field4")] string Field4,' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field5")] DateOnly Field5,' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field6")] DateTime Field6' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field7")] object Field7' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field8")] TeststructureModel Field8' in generated_class
    assert (
        ') : StellaNowMessageBase("test_event", new List<EntityType>{ new EntityType("entity1", entity1Id), new EntityType("entity2", entity2Id) });'
        in generated_class
    )


def test_generate_model_class(model_details):
    model_id = "34a736ed-0b51-415a-ac24-0c6a975e9381"
    model = model_details[model_id]

    # Generate the model class
    generated_model_class = CsharpCodeGenerator.generate_model_class(model, model_details)

    # Assertions
    assert f"public record TeststructureModel" in generated_model_class
    assert '[property: Newtonsoft.Json.JsonProperty("fieldA")] string FieldA,' in generated_model_class
    assert '[property: Newtonsoft.Json.JsonProperty("fieldB")] int FieldB' in generated_model_class


def test_generate_nested_model_class(model_details):
    model_id = "66a744ed-0b32-125a-ac24-0c6a525e9381"
    model = model_details[model_id]

    # Generate the model class
    generated_model_class = CsharpCodeGenerator.generate_model_class(model, model_details)

    # Assertions
    assert "using StellaNowSDK.Messages;" not in generated_model_class
    assert "using StellaNowSDK.Messages.Models;" not in generated_model_class
    assert "namespace StellaNowSDK.Messages.Models;" in generated_model_class
    assert f"public record TestnestedModel(" in generated_model_class
    assert '[property: Newtonsoft.Json.JsonProperty("fieldC")] string FieldC,' in generated_model_class
    assert '[property: Newtonsoft.Json.JsonProperty("fieldD")] TestanotherModel FieldD' in generated_model_class


def test_generate_empty_model_class(test_event_without_model, model_details_empty):
    generated_class = CsharpCodeGenerator.generate_message_class(test_event_without_model, model_details_empty)

    assert "using StellaNowSDK.Messages;" in generated_class
    assert "using StellaNowSDK.Messages.Models;" not in generated_class
    assert "namespace StellaNowSDK.Messages;" in generated_class
    assert "[property: Newtonsoft.Json.JsonIgnore] string entity1Id," in generated_class
    assert "[property: Newtonsoft.Json.JsonIgnore] string entity2Id," in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field1")] decimal Field1,' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field2")] int Field2,' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("fieldValue3")] bool FieldValue3,' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field4")] string Field4,' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field4")] string Field4,' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field5")] DateOnly Field5,' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field6")] DateTime Field6' in generated_class
    assert '[property: Newtonsoft.Json.JsonProperty("field7")] object Field7' in generated_class
    assert (
        ') : StellaNowMessageBase("test_event_no_model", new List<EntityType>{ new EntityType("entity1", entity1Id), new EntityType("entity2", entity2Id) });'
        in generated_class
    )
